import numpy as np

def batch_normalization_forward(x, gamma,beta,params):
    """
    Input:
        x: Array of shape (N,D)
        gamma : Scale Parameter of shape (D,)
        beta : Shift Parameter of shape (D,)
        params:
            'mode' : 'train' or 'test'
            'running_var' : Current value of exponential decayed variance over train set
            'running_mean': Current value of exponential decayed mean over train set
            'momentum' : Rate at which mean and variance should be decayed
            'eps' : Epsilon value to prevent shooting of variance
    """
    mode = params.get('mode','train')
    momentum = params.get('momentum',0.9)
    running_mean = params.get('running_mean',np.zeros(x.shape[1]))
    running_var  = params.get('running_var',np.zeros(x.shape[1]))
    eps = params.get('eps',1e-8)
    
    out, cache = None, None
    if mode == "train":
        mu = np.mean(x,axis=0)
        t  = x - mu
        f  = t**2
        u = np.mean(f,axis=0)
        k = np.power(u+eps,0.5)
        p = 1./k
        q = t*p
        r = q*gamma
        out = r + beta
        params['running_mean']  = momentum*running_mean + (1-momentum)*mu
        params['running_var']   = momentum*running_var + (1-momentum)*u
        cache = (x,gamma,q,t,p)
    else:
        t = x - running_mean
        u = running_var
        k = np.power(u + eps,0.5)
        p = 1/k
        q = t*p
        r = q*gamma
        out = beta + r
        
    return out,cache

def batch_normalization_backward(dOut,cache):
    """
        Input
            dOut : Upstream Gradients of shape (N,D)
            cache : (x,gamma,beta,params,q,t,p,k)
        Output
            dx,dgamma,dbeta of same shape as x,gamma,beta
    """
    x, gamma, q, t, p = cache
    N, D = x.shape
    dbeta  = np.sum(dOut,axis=0)
    dgamma = np.sum(dOut*q,axis=0)
    dr = dOut
    dq = gamma*dOut
    dp = np.sum(t*dq,axis=0)
    dt1 = p*dq
    dk = -(p**2)*dp
    du = 0.5*p*dk
    df = np.ones((N,D))*du*1./N
    dt2 = 2*t*df
    dt = dt1+dt2
    dx1 = dt
    dmu = -np.sum(dx1,axis=0)
    dx2 = 1./N*np.ones((N,D))*dmu
    dx = dx1+dx2
    
    return dx,dgamma,dbeta

def spatial_batch_forward(x,gamma,beta,params):
    """
    Input:
        x: Array of shape (N,C,H,W)
        gamma : Scale Parameter of shape (C,)
        beta : Shift Parameter of shape (C,)
        params:
            'mode' : 'train' or 'test'
            'running_var' : Current value of exponential decayed variance over train set
            'running_mean': Current value of exponential decayed mean over train set
            'momentum' : Rate at which mean and variance should be decayed
            'eps' : Epsilon value to prevent shooting of variance
    """
    N,C,H,W=x.shape
    #Reshapes array to N,H,W,C
    reshaped_x = np.swapaxes(np.swapaxes(x,1,2),2,3)
    reshaped_x = reshaped_x.reshape(-1,C)
    
    out,cache = batch_normalization_forward(reshaped_x,gamma,beta,params)
    out = out.reshape(N,H,W,C)
    out = np.swapaxes(np.swapaxes(out,2,3),2,1)
    return out,cache

def spatial_batch_backward(dOut,cache):
    """
        Input
            dOut : Upstream Gradients of shape (N,C,H,W)
            cache : (x,gamma,beta,params,q,t,p,k)
        Output
            dx,dgamma,dbeta of same shape as x,gamma,beta
    """
    N,C,H,W = dOut.shape
    reshaped_dOut = np.swapaxes(np.swapaxes(dOut,1,2),2,3)
    reshaped_dOut = reshaped_dOut.reshape(-1,C)
    dx,dgamma,dbeta = batch_normalization_backward(reshaped_dOut,cache)
    dx = dx.reshape(N,H,W,C)
    dx = np.swapaxes(np.swapaxes(dx,2,3),1,2)
    return dx,dgamma,dbeta


from layers.layer import *
from operator import mul
import matplotlib.pyplot as plt
import pickle
import json
import jsonpickle
import jsonpickle.ext.numpy as jsonpickle_numpy

jsonpickle_numpy.register_handlers()
class network:
    
    def __init__(self,input_shape,update_params,initialization="normal"):
        self.J=[]
        self.layers=[]
        self.accuracies=[]
        self.out_shape=[input_shape]
        self.update_params=update_params
        self.initialization = initialization
        
    def initializer(self,mean=0,shift=0.01,shape=None,initialization="normal"):
        if initialization == "normal":
            init = shift*np.random.standard_normal(shape)+mean
        elif initialization == "xavier":
            init = np.random.standard_normal(shape)/(shape[0]**0.5)
        elif initialization == "xavier2":
            if len(shape)==2:
                init = np.random.standard_normal(shape)/((shape[0]/2.)**0.5)
            elif len(shape)==4:
                init = np.random.standard_normal(shape)/((reduce(mul,shape[1:],1)/2.)**0.5)
        return init

    def add(self,layer_name,affine_out=None,
            padding_h=None,padding_w=None,
            pooling_params=None,
            num_kernels=None,kernel_h=None,kernel_w=None,convolution_params=None,
            batch_params=None,
            output=None,
            mean=0,shift=0.01):
        
        outshape = len(self.out_shape[-1])
        
        if layer_name == "affine" and outshape==2:
            
            N,D = self.out_shape[-1]
            W = self.initializer(mean,shift,(D,affine_out),initialization=self.initialization)
            b = np.zeros(affine_out,)
            self.layers.append(Affine(W,b,self.update_params))
            self.out_shape.append((N,affine_out))
            
        elif layer_name == "flatten" and outshape>2:
        
            self.layers.append(Flatten())
            shape = self.out_shape[-1]
            self.out_shape.append((shape[0],reduce(mul,shape[1:],1)))
        
        elif layer_name == "relu":
            
            self.layers.append(Relu())
            shape = self.out_shape[-1]
            self.out_shape.append(shape)
        
        elif layer_name == "sigmoid":
            
            self.layers.append(Sigmoid())
            shape = self.out_shape[-1]
            self.out_shape.append(shape)
            
        elif layer_name == "tanh":
            
            self.layers.append(Tanh())
            shape = self.out_shape[-1]
            self.out_shape.append(shape)
            
        elif layer_name == "leaky_relu":
            
            self.layers.append(LeakyRelu())
            shape = self.out_shape[-1]
            self.out_shape.append(shape)
            
        elif layer_name == "padding" and outshape == 4:
            
            shape = self.out_shape[-1]
            self.layers.append(Padding(padding_h,padding_w))
            self.out_shape.append((shape[0],shape[1],2*padding_h+shape[2],2*padding_w+shape[3]))
            
        elif layer_name == "pooling" and outshape == 4:
            
            self.layers.append(Pooling(pooling_params))
            
            Ph = pooling_params.get('pooling_height',2)
            Pw = pooling_params.get('pooling_width',2)
            PSH = pooling_params.get('pooling_stride_height',2)
            PSW = pooling_params.get('pooling_stride_width',2)
    
            N,C,H,W = self.out_shape[-1]
            Hout = (H-Ph)//PSH + 1
            Wout = (W-Pw)//PSW + 1
            
            self.out_shape.append((N,C,Hout,Wout))
    
        elif layer_name == "convolution" and outshape == 4:
            
            N,C,H,W = self.out_shape[-1]
            S = convolution_params.get('stride',1)
            Hout = abs(H-kernel_h)//S + 1
            Wout = abs(W-kernel_w)//S + 1
        
            W = self.initializer(mean,shift,(num_kernels,C,kernel_h,kernel_w),initialization=self.initialization)
            b = np.zeros((num_kernels,))
            self.layers.append(Convolution(W,b,convolution_params,self.update_params))
            self.out_shape.append((N,num_kernels,Hout,Wout))
            
        elif layer_name == "softmax" and outshape==2:
            
            self.layers.append(Softmax())
            self.out_shape.append((1))
            
        elif layer_name == "svm" and outshape==2:
            
            self.layers.append(SVM())
            self.out_shape.append((1))
            
        elif layer_name == "mse" and outshape == 2 and self.out_shape[-1][1]==1:
            
            self.layers.append(MSE())
            self.out_shape.append((1))
            
        elif layer_name == "batch_normalization" and outshape==2:
            
            shape = self.out_shape[-1]
            D = shape[1]
            
            gamma = np.ones((D,))
            beta = np.zeros((D,))
            self.layers.append(BatchNormalization(gamma,beta,batch_params,self.update_params))
            self.out_shape.append(shape)
        elif layer_name == "spatial_batch" and outshape==4:
            
            shape = self.out_shape[-1]
            gamma = np.ones((shape[1],))
            beta = np.zeros((shape[1],))
            self.layers.append(SpatialBatchNormalization(gamma,beta,batch_params,self.update_params))
            self.out_shape.append(shape)
        else:
            print "Check Shapes"
            raise NotImplementedError
    
    
    def train(self,X,y):
        self.test(X[:64],y[:64])
        print("Initial Cost :"+str(self.J[-1]))
        print("Initial Accuracy :"+str(self.accuracies[-1]))
        
        for i in range(self.update_params['epoch']):
            sample  = np.random.randint(0,X.shape[0],(self.out_shape[0][0],))
            inp = X[sample]
            loss = 0.0
            for layer in self.layers[:-1]:
                inp = layer.forward(inp)
                loss += layer.loss_reg()
            
            scores,inp = self.layers[-1].forward(inp,y[sample])
            
            for layer in self.layers[::-1]:
                inp = layer.backprop(inp)
            
            self.test(X[sample],y[sample])
            print("Cost at Iteration "+str(i)+" : "+str(self.J[-1]))
            print("Accuracy at Iteration "+str(i)+" : "+str(self.accuracies[-1]))
                  
    def test(self,X,y, accuracies = None,costs= None):
        if accuracies is None:
            accuracies = self.accuracies
            costs = self.J
        loss = 0.0
        inp = X
        for layer in self.layers[:-1]:
            inp = layer.forward(inp)
            loss += layer.loss_reg()
        
        scores,inp = self.layers[-1].forward(inp,y)
        accuracies.append(self.accuracy(scores,y))
        costs.append(inp+loss)
    
    def batch_test(self,X,y):
        accuracies = []
        costs = []
        batch_size = self.out_shape[0][0]
        for i in range(0,X.shape[0],batch_size):
            end = min(i+batch_size,X.shape[0])
            self.test(X[i:end],y[i:end],accuracies,costs)
    
        accuracies = np.array(accuracies)
        costs = np.array(costs)
        print("Accuracy : ",np.mean(accuracies))
        print("Cost : ",np.sum(costs))
    
    def accuracy(self,scores,y):
        return 1.0*np.sum(np.argmax(scores,axis=1)==y)/y.shape[0]
    
    def predict(self,X):
        inp = X
        for layer in self.layers:
            inp = layer.forward(inp)
            
        return np.argmax(inp,axis=1)
    
    def save(self,filename):
        outfile = open('models/'+filename, 'wb')
        enc = jsonpickle.encode(self)
        json.dump(enc,outfile)
    
    @staticmethod
    def load(filename):
        infile = open('models/'+filename,'rb')
        return jsonpickle.decode(json.load(infile))
    
    def plot(self):
        plt.plot(self.J[1:])
        plt.ylabel("Loss")
        plt.xlabel("Iterations")
        plt.title("Loss VS Iteration")
        plt.show()



import numpy as np
def rel_error(x, y):
    """ returns relative error """
    return np.max(np.abs(x - y) / (np.maximum(1e-8, np.abs(x) + np.abs(y))))


def num_gradient_array(f,x,dOut,h=1e-5):
    grad = np.zeros_like(x)
    it = np.nditer(x, flags=['multi_index'], op_flags=['readwrite'])
    while not it.finished:
        ix = it.multi_index
        x[ix]+=h
        fxph=f(x).copy()
        x[ix]-=2*h
        fxmh=f(x).copy()
        x[ix]+=h
        grad[ix] = np.sum((fxph-fxmh)*dOut)/(2*h)
        it.iternext()

    return grad



from util.layer import *
from util.util import *
from util.activations import *
from util.normalization import *
from util.convolution import *
from util.loss import *
from descent.descent import *
import numpy as np
import copy

def regularization(weight,params):
    reg = params.get('reg',0.0)
    reg_type = params.get('reg_type')
    #print(str(reg) + " :  "+reg_type)
    if reg_type is not 'None':
        reg = params.get('reg',0.0)
        if reg_type is 'L2':
            return reg*np.sum(np.square(weight))
        elif reg_type is 'L1':
            return reg*np.sum(np.abs(weight))
        else:
            return 0.0
    else:
        return 0.0
    
class Padding():
    
    def __init__(self,padding_h,padding_w):
        self.padding = (padding_h,padding_w)
    
    def forward(self,X):
        return padding_forward(X,self.padding[0],self.padding[1])
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut):
        return padding_backward(dOut,self.padding)
    
class Pooling():
    
    def __init__(self,pooling_params):
        self.pooling_params = pooling_params
        self.cache = None
        
    def forward(self,X):
        out,self.cache = max_pooling_forward(X,self.pooling_params)
        return out
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut):
        out =  max_pooling_backward(dOut,self.cache)
        self.cache = None
        return out

class Convolution():
    
    def __init__(self,w,b,convolution_params,update_params):
        self.w = w
        self.b = b
        self.convolution_params = convolution_params
        self.wparams=copy.deepcopy(update_params)
        self.bparams=copy.deepcopy(update_params)
        
    def forward(self,X):
        out,self.cache = convolve_forward_fast(X,self.w,self.b,self.convolution_params)
        return out
    
    def loss_reg(self):
        return regularization(self.w,self.wparams)
    
    def backprop(self,dOut):
        dx,dw,db = convolve_backward_fast(dOut,self.cache)
        update_weight(self.w,dw,self.wparams,regularization=True)
        update_weight(self.b,db,self.bparams)
        self.cache = None
        return dx

class Relu():
    
    def __init__(self):
        self.cache = None
    
    def forward(self,X):
        out,self.cache = relu_forward(X)
        return out
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut):
        out= relu_backward(dOut,self.cache)
        self.cache = None
        return out
    
class Sigmoid():
    
    def __init__(self):
        self.cache = None
    
    def forward(self,X):
        out,self.cache = sigmoid_forward(X)
        return out
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut):
        out= sigmoid_backward(dOut,self.cache)
        self.cache = None
        return out
    
class Tanh():
    
    def __init__(self):
        self.cache = None
    
    def forward(self,X):
        out,self.cache = tanh_forward(X)
        return out
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut):
        out = tanh_backward(dOut,self.cache)
        self.cache = None
        return out

class LeakyRelu():
    
    def __init__(self):
        self.cache = None
    
    def forward(self,X):
        out,self.cache = leaky_relu_forward(X)
        return out
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut):
        out = leaky_relu_backward(dOut,self.cache)
        self.cache = None
        return out
    
class Affine():
    
    def __init__(self,w,b,update_params):
        self.cache = None
        self.w = w
        self.b = b
        self.wparams = copy.deepcopy(update_params)
        self.bparams = copy.deepcopy(update_params)
        
    def forward(self,X):
        out, self.cache = affine_forward(X,self.w,self.b)
        return out
    
    def loss_reg(self):
        return regularization(self.w,self.wparams)
    
    def backprop(self,dOut):
        dx,dw,db = affine_backward(dOut,self.cache)
        update_weight(self.w,dw,self.wparams,regularization=True)
        update_weight(self.b,db,self.bparams)
        self.cache = None
        return dx

class Flatten():
    
    def __init__(self):
        self.cache = None
    
    def forward(self,X):
        out,self.cache = flatten_forward(X)
        return out
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut):
        out= flatten_backward(dOut,self.cache)
        self.cache = None
        return out
    

class Softmax():
    
    def __init__(self):
        self.dx= None
    
    def forward(self,X,y=None):
        if y is None:
            scores = softmax_loss(X)
            return scores
        else:
            scores,loss,self.dx = softmax_loss(X,y)
            return scores,loss
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut=None):
        return self.dx

class SVM():

    def __init__(self):
        self.dx= None
    
    def forward(self,X,y=None):
        if y is None:
            scores = softmax_loss(X)
            return scores
        else:
            scores,loss,self.dx = svm_loss(X,y)
            return scores,loss
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut=None):
        return self.dx

def MSE():
    
    def __init__(self):
        self.dx = None
    
    def forward(self,X,y=None):
        if y is None:
            scores = mse_loss(X)
            return scores
        else:
            scores,loss,self.dx = mse_loss(X,y)
            return scores,loss
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut=None):
        return self.dx
    
class BatchNormalization():
    
    def __init__(self,gamma,beta,params,update_params):
        self.gamma = gamma
        self.beta = beta
        self.params = params
        self.gamma_update_params = copy.deepcopy(update_params)
        self.beta_update_params  = copy.deepcopy(update_params)
        self.cache = None
        
    def forward(self,X):
        out,self.cache = batch_normalization_forward(X,self.gamma,self.beta,self.params)
        return out
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut):
        dx,dgamma,dbeta = batch_normalization_backward(dOut,self.cache)
        update_weight(self.gamma,dgamma,self.gamma_update_params)
        update_weight(self.beta,dbeta,self.beta_update_params)
        self.cache = None
        return dx

class SpatialBatchNormalization():
    
    def __init__(self,gamma,beta,params,update_params):
        self.gamma = gamma
        self.beta = beta
        self.params = params
        self.gamma_update_params = copy.deepcopy(update_params)
        self.beta_update_params = copy.deepcopy(update_params)
        self.cache = None
    
    def forward(self,X):
        out,self.cache = spatial_batch_forward(X,self.gamma,self.beta,self.params)
        return out
    
    def loss_reg(self):
        return 0.0
    
    def backprop(self,dOut):
        dx,dgamma,dbeta = spatial_batch_backward(dOut,self.cache)
        update_weight(self.gamma,dgamma,self.gamma_update_params)
        update_weight(self.beta,dbeta,self.beta_update_params)
        self.cache = None
        return dx        



import numpy as np

def update_weight(W,dW,params,regularization=False):
    """
        Input:
            W: Initial Weights
            dW: Gradient wrt Weights
            params: For Gradient Descent Method
           'alpha':learning_rate
           'reg' : regularization constant
           'reg_type':'None/L1'/'L2'
           'method': 'gd' | 'momentum' | 'adagrad'  |'rmsprop' | 'partial_adam' | 'adam'
           'epoch': Number of times to run gradient descent
           'momentum':  momentum (In momentum)
           'velocity' : gradient velocity for momentum method (In momentum)
           'grad_square' : sum of square of gradients (In rmsprop && adagrad)
           'offset'      : offset (about 1e-7 to avoid /0) (In rmsprop && adagrad && partial_adam && adam)
           'decay'       : value to decay grad_square by in rmsprop (In rmsprop)
           'beta1'       : decay value for firstmoment (In adam && partial_adam)
           'beta2'       : decay value for secondmoment (In adam && partial_adam)
           'firstmoment' : firstmoment value (In adam && partial_adam)
           'secondmoment': secondmoment value (In adam && partial_adam)
           't'           : update number (Only in Adam)
    """
    alpha = params.get('alpha',0)
    method = params.get('method',"gd")
    reg_type = params.get('reg_type','None')
    if reg_type is not 'None':
        reg = params.get('reg',0.0)
        if reg_type is 'L2':
            dW += 2*reg*W
        elif reg_type is 'L1':
            dW +=2*reg
    
    if method == "gd":
    
        W -= alpha*dW
    
    elif method == "momentum":
        
        momentum  = params.get("momentum",0.99)
        v = params.get("velocity",0.0)
        
        v = momentum*v + alpha*dW
        W -= v
        
        params["velocity"]=v
        
    elif method == "adagrad":
        
        gradSquare = params.get("grad_square",0.0)
        offset = params.get("offset",1e-7)
        
        gradSquare += np.square(dW)
        W -=alpha*dW/(np.sqrt(gradSquare)+offset)
        
        params["grad_square"]=gradSquare
    
    elif method == "rmsprop":
        
        decay = params.get("decay",0.9)
        gradSquare = params.get("grad_square",0.0)
        offset = params.get("offset",1e-7)
        
        gradSquare = decay*gradSquare + (1-decay)*np.square(dW)
        W -= alpha*dW/(np.sqrt(gradSquare)+offset)
        
        params["grad_square"]=gradSquare
    elif method == "partial_adam":
        
        beta1 = params.get("beta1",0.9)
        beta2 = params.get("beta2",0.999)
        mu1 = params.get("firstmoment",0.0)
        mu2 = params.get("secondmoment",0.0)
        offset = params.get("offset",1e-7)
        
        mu1 = beta1*mu1 + (1-beta1)*dW
        mu2 = beta2*mu2 + (1-beta2)*np.square(dW)
        W -= alpha*mu1/(np.sqrt(mu2)+offset)
        
        params["firstmoment"]=mu1
        params["secondmoment"]=mu2
        
    elif method == "adam":
        
        beta1 = params.get("beta1",0.9)
        beta2 = params.get("beta2",0.999)
        mu1 = params.get("firstmoment",0.0)
        mu2 = params.get("secondmoment",0.0)
        offset = params.get("offset",1e-7)
        t = params.get("t",0)
        t+=1
        mu1 = beta1*mu1 + (1-beta1)*dW
        mu2 = beta2*mu2 + (1-beta2)*np.square(dW)
        unbias1 = mu1/(1-beta1**t)
        unbias2 = mu2/(1-beta2**t)
        W -= alpha*unbias1/(np.sqrt(unbias2)+offset)
        
        
        params["firstmoment"]=mu1
        params["secondmoment"]=mu2
        params["t"]=t
    else:
        raise NotImplementedError


import numpy as np
from scipy.stats import threshold
def relu_forward(x):
    """
    Input : x of any shape
    Output: Of same shape as x after applying Relu Non Linearity
    """
    out = np.maximum(0,x)
    cache = (x)
    return out,cache

def relu_backward(dOut,cache):
    """
    Input: dOut of any shape is the upstream gradient
    Output:dx of same shape as x
    """
    x = cache
    dx = np.sign(np.maximum(0,x))*dOut
    return dx

def sigmoid_forward(x):
    """
    Input x of any shape

    Output: Sigmoided x
    """
    out = 1+np.exp(-x)
    out = 1/out
    cache = (x,out)
    return out,cache

def sigmoid_backward(dOut,cache):
    """
    Input:
    dOut of any shape

    Output:
    dx of same shape as dOut
    """
    _, out = cache
    dx = out*(1-out)*dOut
    return dx

def tanh_forward(x):
    """
    Input:
        x of any shape
    Output:
        tanh(x) of same shape as x
        cache containing x
    """
    out = np.tanh(x)
    cache = (out)
    return out,cache

def tanh_backward(dOut,cache):
    """
    Input:
        dOut of same shape as x
        cache containing x
        
    Output:
        dx of same shape as x
    """
    
    (out) = cache
    dx = (1-out*out)*dOut
    return dx

def leaky_relu_forward(x):
    """
    Input:
        x of any shape
    Output:
        out:max(0.1x,x) of same shape as x
        cache:x,out
    """
    
    out = np.maximum(0.1*x,x)
    cache = (x)
    return out,cache

def leaky_relu_backward(dOut,cache):
    """
    Input:
        dOut of any shape
        cache:(x,out)
        x is the input given for forward
        out is the leaky_relu_forwarded output
    Output:
        dx of same shape as x
    """
    (x) = cache
    x = np.sign(x)
    dx = threshold(x,threshmin=0,newval=0.1)*dOut
    return dx


from distutils.core import setup

setup(
    name = 'networks',
    packages = ['networks','networks/layers','networks/layers/util/','networks/layers/descent/'],
    version = '0.3.2',  # Ideally should be same as your GitHub release tag varsion
    description = 'Allows to create NN models',
    author = 'Shharrnam Chhatpar',
    author_email = 'sharnam19.nc@gmail.com',
    url = 'https://github.com/sharnam19/Networks',
    download_url = 'https://github.com/sharnam19/Networks/archive/v0.3.3.tar.gz',
    keywords = ['machine-learning', 'deep-learning','neural-networks','linear regression',
                'logistic regression'],
    classifiers = [],
    install_requires=[
        'jsonpickle',
        'numpy',
        'scipy',
        'matplotlib'
    ]
)



import numpy as np
from scipy import signal

def affine_forward(x, w, b):
    """
        Input
        x: Input of shape (N,D)
        w: Weights of shape (D,H)
        b: Biases of shape (H,)

        Output(out,cache)
        Returns out of shape (N,H)

    """
    out = x.dot(w)+b
    cache = (x,w,b)
    return out,cache

def affine_backward(dOut,cache):
    """
        Input:
        dOut: Upstream gradient of shape (N,H)
        Returns Derivative wrt the inputs x,w,b i.e
        dx,dw,db
    """
    x,w,b = cache
    dx = np.dot(dOut,w.T)
    dw = np.dot(x.T,dOut)
    db = np.sum(dOut,axis=0)
        
    return dx,dw,db

def flatten_forward(x):
    """
        Input:
            x of any shape (N,D1,D2,D3,.....Dn)
        Output:
            out : of shape (N,D1*D2*D3*....Dn)
            cache: consisting of shape
    """
    shape = x.shape
    out = x.reshape(shape[0],-1)
    cache = (shape)
    return out,cache

def flatten_backward(dOut,cache):
    """
    Input:
        dOut: Upstream gradients in shape (N,D1*D2*D3*.....Dn)
        cache: (shape of input tensor)
    Output:
        dx: dOut reshaped in shape of x
    """
    (shape)=cache
    dx = dOut.reshape(shape)
    return dx


import numpy as np

def padding_forward(x,Ph,Pv):
    """
        Input:
            x : of any shape
            Ph: Num of horizontal Padding on top and bottom
            Pv: Num of vertical Padding on sides
            
        Output:
            out : Padded input with Ph and Pv layers of padding
            cache:(Ph,Pv)
    """
    shapes = list(x.shape)
    shapes[-2]+=2*Ph
    shapes[-1]+=2*Pv
    
    out = np.zeros(shapes)
    out[:,:,Ph:-Ph,Pv:-Pv]=x
    return out

def padding_backward(dOut,cache):
    """
        Input:
            dOut: any 4D numpy array i.e the upstream gradient
            cache: Contains Horizontal and Vertical Padding Num
        Output:
            dx : of same shape as x
    """
    Ph, Pv = cache
    dx = dOut[:,:,Ph:-Ph,Pv:-Pv]
    return dx

def max_pooling_forward(x,pooling_params):
    """
        Input:
            x: Image of shape (N,C,H,W)
            pooling_params:
                'pooling_height':Pooling Height
                'pooling_width':Pooling Width
                'pooling_stride_height':Pooling Stride Height
                'pooling_stride_width':Pooling Stride Width
        Output:
            out: Pooled Image (N,C,Hout,Wout)
            cache : (out,x,pooling_params)
    """
    Ph = pooling_params.get('pooling_height',2)
    Pw = pooling_params.get('pooling_width',2)
    PSH = pooling_params.get('pooling_stride_height',2)
    PSW = pooling_params.get('pooling_stride_width',2)
    
    N,C,H,W = x.shape
    Hout = (H-Ph)//PSH + 1
    Wout = (W-Pw)//PSW + 1
    out = np.zeros((N,C,Hout,Wout))
    for h in range(Hout):
        top = h*PSH
        bottom = top + Ph
        for w in range(Wout):
            left = w*PSW
            right = left + Pw
            out[:,:,h,w] = np.max(x[:,:,top:bottom,left:right],axis=(2,3))
    
    cache = (out,x,pooling_params)
    return out,cache

def max_pooling_backward(dOut,cache):
    """
        Input:
            dOut: Upstream Gradient
            cache : (out,x,pooling_params)
        Output:
            dx:Gradient wrt x
    """
    
    out,x,pooling_params = cache
    
    dx=np.zeros_like(x)
    
    Ph = pooling_params.get('pooling_height',2)
    Pw = pooling_params.get('pooling_width',2)
    PSH = pooling_params.get('pooling_stride_height',2)
    PSW = pooling_params.get('pooling_stride_width',2)
    
    N,C,Hout,Wout = dOut.shape
    temp_ones = np.ones((N,C,Ph,Pw))
    for h in range(Hout):
        top = h*PSH
        bottom = top+Ph
        for w in range(Wout):
            left = w*PSW
            right = left+Pw
            mask  = temp_ones*out[:,:,h,w][:,:,np.newaxis,np.newaxis]
            max_to_1 = x[:,:,top:bottom,left:right]==mask
            dx[:,:,top:bottom,left:right]=max_to_1*dOut[:,:,h,w][:,:,np.newaxis,np.newaxis]
    
    return dx

def convolve_forward_naive(x,w,b,params):
    """
        Input:
            x: of shape N,C,H,W
            w: of shape D,C,HH,WW
            b: of shape D,
            params:
                'stride' of input
        Output:
            out : convolved Input
            cache : (x,w,b)
    """
    N,C,H,W = x.shape
    D,_,HH,WW = w.shape
    S = params.get('stride',1)
    
    Hout = (H-HH)//S + 1
    Wout = (W-WW)//S + 1
    
    out = np.zeros((N,D,Hout,Wout))
    reshaped_w = w.reshape(D,-1)
    reshaped_w = np.swapaxes(reshaped_w,0,1)
    
    for h in range(Hout):
        top = h*S
        bottom = top + HH
        for w in range(Wout):
            left = w*S
            right = left + WW
            reshaped_x = x[:,:,top:bottom,left:right].reshape(N,-1)
            out[:,:,h,w] = reshaped_x.dot(reshaped_w)
    
    cache = (x,w,params)
    out += b[np.newaxis,:,np.newaxis,np.newaxis]
    
    return out,cache

def convolve(x,w,params,mode='valid',backprop=False):
    """
        Helper Method to calculate Convolution
    """
    
    N,C,H,W = x.shape
    D,_,HH,WW = w.shape
    S = params.get('stride',1)
    
    Hout = abs(H-HH)//S + 1
    Wout = abs(W-WW)//S + 1
    
    out = np.zeros((N,D,Hout,Wout))
    if not backprop:
        w = np.flip(np.flip(w,3),2)
        
    Hpadded = H + HH - 1
    Wpadded = W + WW - 1
    fx = np.fft.fft2(x,(Hpadded,Wpadded),axes=(-2,-1))
    fw = np.fft.fft2(w,(Hpadded,Wpadded),axes=(-2,-1))
    
    padH = (Hpadded-Hout)//2
    padW = (Wpadded-Wout)//2
    fz = fx[:,np.newaxis,:,:,:] * fw[np.newaxis,:,:,:,:]

    fz = np.sum(fz,axis=(2))
    out = np.fft.ifft2(fz)
    if mode == 'valid':
        out = out[:,:,padH:-padH,padW:-padW].real
    elif mode == 'full':
        out = out.real
    return out

def convolve_forward_fast(x,w,b,params):
    """
        Input:
            x: of shape N,C,H,W
            w: of shape D,C,HH,WW
            b: of shape D,
            params:
                'stride':Stride of weight
        Output:
            out : convolved Input
            cache : (x,w,params)
    """
    out = convolve(x,w,params)
    out += b[np.newaxis,:,np.newaxis,np.newaxis]
    
    cache = (x,w,params)
    
    return out,cache

def convolve_backward_fast(dOut,cache):
    """
        Input:
            dOut: Upstream Gradients
            cache : x,w,b
        Output:
            dx:Derivative wrt x
            dw:Derivative wrt w
            db:Derivative wrt b
    """
    x,w,params=cache
    
    N,C,H,W = x.shape
    D,_,HH,WW = w.shape
    
    db = np.sum(dOut,axis=(0,2,3))
    dx = np.zeros_like(x)
    dw = np.zeros_like(w)
    
    shifted_w = np.swapaxes(w,0,1)
    shifted_d = np.swapaxes(dOut,0,1)
    shifted_d = np.flip((np.flip(shifted_d,3)),2)
    shifted_x = np.swapaxes(x,0,1)
    
    dx = convolve(dOut,shifted_w,params,mode='full',backprop=True)
    dw = convolve(shifted_d,shifted_x,params,mode='valid',backprop=True)
    
    return dx,dw,db



import numpy as np
import copy
def softmax_loss(x,y=None):
    """
        Input:
            x of shape (N,D)
            y of shape (N,). It is the class values
        Output:
            loss: Loss Value
            dx: Softmax Loss wrt the input. Same Shape as x
    """
    maximum = np.max(x,axis=1)
    shifted_x = x - maximum[:,np.newaxis]
    exp_x = np.exp(shifted_x)
    scores = exp_x/np.sum(exp_x,axis=1)[:,np.newaxis]
    N,D = x.shape
    if y is None:
        return scores
    loss = 0
    loss += np.sum(-np.log(scores[range(N),y]))/N
    
    offset = np.zeros_like(scores)
    offset[range(N),y]=1
    dx = (scores-offset)/N
    return scores,loss,dx

def svm_loss(x,y=None):
    """
        Input:
            x of shape (N,D)
            y of shape (N,). It is the class values
        Output:
            loss : loss value
            dx  : SVM Loss wrt input. Same shape as x
    """
    scores = x
    toRet = copy.deepcopy(scores)
    if y is None:
        return scores
    
    N,D = x.shape
    correct_scores = scores[range(N),y][:,np.newaxis]
    margin = scores - correct_scores + 1
    scores[range(N),y]=0
    scores = np.maximum(0,scores)
    loss = np.sum(scores)/N
    
    ones = np.sign(scores)
    row_sum = np.sum(ones,axis=1)
    ones[range(N),y]=-row_sum
    dx = ones/N
    return toRet,loss,dx

def mse_loss(x,y=None):
    """
        Input:
            x of shape (N,) 
            y of shape (N,). It is continuous values
        Output:
            loss : loss value
            dx :RMSE Loss wrt input. Same shape as x
    """
    scores = x
    if y is None:
        return scores
    
    N = x.shape[0]
    diff = x-y
    loss = np.mean(np.square(diff))/2
    dx = diff
    return x,loss,dx